from manm_cs.noise.noise import Noise
from manm_cs.prob_distributions.discrete import DiscreteDistribution


class DiscreteNoise(Noise):

    def __init__(self, prob_distribution: DiscreteDistribution):
        super(DiscreteNoise, self).__init__(prob_distribution=prob_distribution)

    def get_num_values(self):
        return self.prob_distribution.get_num_values()



import argparse
import math
from typing import Type, Callable, Optional, Any

import networkx as nx

from manm_cs.graph import Graph, GraphBuilder
from manm_cs.utils import write_single_csv

GROUND_TRUTH_FILE = "ground_truth"
SAMPLES_FILE = "samples"


def type_in_range(type_: Type, lower_bound: Optional[float], upper_bound: Optional[float]) -> \
        Callable[[str], Any]:
    def assert_float_in_range(x: str):
        try:
            x = type_(x)
        except ValueError:
            raise argparse.ArgumentTypeError(f"Cannot convert {x} to type {type_}")

        if lower_bound is not None and x < lower_bound:
            raise argparse.ArgumentTypeError(f"{x} not in range [{lower_bound}, {upper_bound}]")
        if upper_bound is not None and x > upper_bound:
            raise argparse.ArgumentTypeError(f"{x} not in range [{lower_bound}, {upper_bound}]")
        return x

    return assert_float_in_range

def type_in_range_exclusive(type_: Type, lower_bound: Optional[float], upper_bound: Optional[float]) -> \
        Callable[[str], Any]:
    def assert_float_in_range(x: str):
        try:
            x = type_(x)
        except ValueError:
            raise argparse.ArgumentTypeError(f"Cannot convert {x} to type {type_}")

        if lower_bound is not None and x <= lower_bound:
            raise argparse.ArgumentTypeError(f"{x} not in range ({lower_bound}, {upper_bound})")
        if upper_bound is not None and x > upper_bound:
            raise argparse.ArgumentTypeError(f"{x} not in range ({lower_bound}, {upper_bound})")
        return x

    return assert_float_in_range

def to_bool(bool_str: str):
    return bool(int(bool_str))

def lin_func(x):
    return x

def quad_func(x):
    return math.pow(x,2)

def cube_func(x):
    return math.pow(x,3)

FUNCTION_DICTIONARY = {'linear':lin_func,
                       'quadratic':quad_func,
                       'cubic':cube_func,
                       'tanh':math.tanh,
                       'sin':math.sin,
                       'cos':math.cos}

def funcs(function_str: str):
    try:
        val,func = function_str.split(',')
        return(float(val),FUNCTION_DICTIONARY[func])
    except:
        raise argparse.ArgumentTypeError(f"{function_str} has wrong format or not supported")


def parse_args():
    parser = argparse.ArgumentParser(
        description='Generate a dataset for benchmarking causal structure learning using '
                    'the mixed additive noise model')
    group1 = parser.add_mutually_exclusive_group(required=True)
    group1.add_argument('--num_nodes', type=type_in_range(int, 1, None),
                        help='Defines the number of nodes to be in the generated DAG.')
    arg_nx_file = group1.add_argument('--graph_structure_file', type=str, required=False,
                    help='valid .gml file to load a fixed graph structure to networkx.DiGraph structure.')
    group2 = parser.add_mutually_exclusive_group(required=True)
    group2.add_argument('--edge_density', type=type_in_range(float, 0.0, 1.0),
                        help='Defines the density of edges in the generated DAG.')
    group2._group_actions.append(arg_nx_file)
    parser.add_argument('--discrete_node_ratio', type=type_in_range(float, 0.0, 1.0), required=True,
                        help='Defines the percentage of nodes that shall be of discrete type. Depending on its value '
                             'the appropriate model (multivariate normal, mixed gaussian, discrete only) is chosen.')
    parser.add_argument('--num_samples', type=type_in_range(int, 1, None), required=True,
                        help='Defines the number of samples that shall be generated from the DAG.')
    parser.add_argument('--discrete_signal_to_noise_ratio', type=type_in_range(float, 0.0, 1.0),
                        required=False, default=0.9,
                        help='Defines the probability that no noise is added within the mixed additive noise model.')
    parser.add_argument('--min_discrete_value_classes', type=type_in_range(int, 2, None), default=3,
                        required=False,
                        help='Defines the minimum number of discrete classes a discrete variable shall have.')
    parser.add_argument('--max_discrete_value_classes', type=type_in_range(int, 2, None), default=4,
                        required=False,
                        help='Defines the maximum number of discrete classes a discrete variable shall have.')
    parser.add_argument('--continuous_noise_std', type=type_in_range(float, 0.0, None),
                        required=False, default=1.0,
                        help='Defines the standard deviation of gaussian noise added to continuous variables.')
    parser.add_argument('--num_processes', type=type_in_range(int, 1, None), required=False, default=1,
                        help='Defines the number of processes used to sample data from the created graph.')
    parser.add_argument('--functions', type=funcs, required=False, default=[(1.0,lin_func)], nargs='*',
                        help='Defines the probability and functions for relationships between continuous variables '
                             'probabilities have to sum up to 1, supported functions are '
                             'linear, quadratic, cubic, tanh, sin, cos '
                             'format is probabilityF1,F1 probabilityF2,F2 ... .')
    parser.add_argument('--conditional_gaussian', type=to_bool, required=False, default=True,
                        help='Defines if conditional gaussian model is assumed for a mixture of variables. '
                             'possible values are 0 for False and 1 for True.')
    parser.add_argument('--beta_lower_limit', type=type_in_range_exclusive(float, 0.0, None),
                        required=False, default=0.5,
                        help='Defines the lower limit for beta values used for continuous parents. '
                        'Should be greater than 0 and smaller than upper_limit. Note that we sample from the union of [-upper,-lower] and [lower,upper]')
    parser.add_argument('--beta_upper_limit', type=type_in_range_exclusive(float, 0.0, None),
                        required=False, default=1.0,
                        help='Defines the upper limit for beta values used for continuous parents. '
                        'Should be larger than lower_limit. Note that we sample from the union of [-upper,-lower] and [lower,upper]')
    parser.add_argument('--output_ground_truth_file', type=str, required=False, default=GROUND_TRUTH_FILE,
                    help='Output file (path) for the generated ground truth graph (gml). Relative to the directory from which the library is executed. '
                    'Specify without file extension.')
    parser.add_argument('--output_samples_file', type=str, required=False, default=SAMPLES_FILE,
                    help='Output file (path) for the generated samples csv. Relative to the directory from which the library is executed.'
                    'Specify without file extension.')
    args = parser.parse_args()

    assert args.min_discrete_value_classes <= args.max_discrete_value_classes, \
        f"Expected min_discrete_value_classes <= max_discrete_value_classes but got min: " \
        f"{args.min_discrete_value_classes}, max: {args.max_discrete_value_classes} "

    return args


# python src --num_nodes=5 --edge_density=0.6 --discrete_node_ratio=0.0 --num_samples=1000 --discrete_signal_to_noise_ratio=0.0 --min_discrete_value_classes=2 --max_discrete_value_classes=3 --continuous_noise_std=0.2 --functions 0.7,linear 0.3,quadratic --conditional_gaussian 0


def graph_from_args(args) -> Graph:
    if args.graph_structure_file:
        dag = nx.read_gml(args.graph_structure_file)
        assert nx.is_directed_acyclic_graph(dag)
        return GraphBuilder() \
            .with_networkx_DiGraph(dag) \
            .with_discrete_node_ratio(args.discrete_node_ratio) \
            .with_discrete_signal_to_noise_ratio(args.discrete_signal_to_noise_ratio) \
            .with_min_discrete_value_classes(args.min_discrete_value_classes) \
            .with_max_discrete_value_classes(args.max_discrete_value_classes) \
            .with_continuous_noise_std(args.continuous_noise_std) \
            .with_functions(args.functions) \
            .with_conditional_gaussian(args.conditional_gaussian) \
            .with_betas(args.beta_lower_limit, args.beta_upper_limit) \
            .build()
    else:
        return GraphBuilder() \
            .with_num_nodes(args.num_nodes) \
            .with_edge_density(args.edge_density) \
            .with_discrete_node_ratio(args.discrete_node_ratio) \
            .with_discrete_signal_to_noise_ratio(args.discrete_signal_to_noise_ratio) \
            .with_min_discrete_value_classes(args.min_discrete_value_classes) \
            .with_max_discrete_value_classes(args.max_discrete_value_classes) \
            .with_continuous_noise_std(args.continuous_noise_std) \
            .with_functions(args.functions) \
            .with_conditional_gaussian(args.conditional_gaussian) \
            .with_betas(args.beta_lower_limit, args.beta_upper_limit) \
            .build()


if __name__ == '__main__':
    args = parse_args()
    graph = graph_from_args(args)

    dfs = graph.sample(num_observations=args.num_samples, num_processes=args.num_processes)
    write_single_csv(dataframes=dfs, target_path=f"{args.output_samples_file}.csv")

    nx_graph = graph.to_networkx_graph()
    nx.write_gml(nx_graph, f"{args.output_ground_truth_file}.gml")



from abc import ABC

from manm_cs.prob_distributions.prob_distribution import ProbDistribution
from manm_cs.variables import VariableType


class ContinuousDistribution(ProbDistribution, ABC):
    type = VariableType.CONTINUOUS



from typing import List, Optional, Callable, Dict, Tuple

import numpy as np
import pandas as pd

from manm_cs.noise.continous_noise import ContinuousNoise
from manm_cs.variables.variable import Variable, VariableType


class ContinuousVariable(Variable):
    type = VariableType.CONTINUOUS
    continuous_mapper_func: Callable[..., float]
    def __init__(self, idx: int, noise: ContinuousNoise, parents: Optional[List[Variable]] = None,
                 functions: Optional[List[Callable[..., float]]] = None, betas: Optional[List[float]] = None):
        parents = [] if parents is None else parents
        functions = [] if functions is None else functions
        super(ContinuousVariable, self).__init__(idx=idx, parents=parents, noise=noise, functions=functions)
        self.betas = betas
        self.continuous_mapper_func = self.__create_continuous_mapper_func(functions=functions, betas=betas)

        # Input parameter validation
        if len(functions) != len(self._get_continuous_parents()):
            raise ValueError(f'There must be one function for each continuous parent. Expected '
                             f'{len(self._get_continuous_parents())}, but were {len(functions)}')

    def __create_continuous_mapper_func(self, functions: List[Callable[..., float]], betas: List[float]) -> Callable[[pd.Series], float]:
        """Creates the mapper function for continuous parents

        """

        def func(parent_values: pd.Series):
            return np.sum([betas[i] * functions[i](value) for i, value in enumerate(parent_values)])

        return func

    def get_non_root_signal(self, df: pd.DataFrame) -> pd.Series:
        # Compute signal for continuous parent variables
        continuous_parent_idxs = [p.idx for p in self._get_continuous_parents()]
        continuous_signal = df[continuous_parent_idxs].apply(self.continuous_mapper_func, axis=1) \
            if len(continuous_parent_idxs) > 0 \
            else pd.Series(np.zeros(len(df)))

        # Compute signal for discrete parent variables
        discrete_parent_idxs = [p.idx for p in self._get_discrete_parents()]
        discrete_signal = df[discrete_parent_idxs].apply(sum, axis=1) \
            if len(discrete_parent_idxs) > 0 \
            else pd.Series(np.zeros(len(df)))

        # Aggregate continuous and discrete signal terms into overall signal term
        return continuous_signal + discrete_signal

    def sample(self, df: pd.DataFrame, num_observations: int) -> pd.Series:
        if self._is_root():
            # If the variable is a root variable, the sampling is determined by the noise term only
            signal = pd.Series(np.zeros(num_observations, dtype=float))
        else:
            # If the variable has one or more parent variables, the sampling is driven
            # by a combination of signal and noise term
            signal = self.get_non_root_signal(df=df)

        # Add noise and signal terms
        return self.noise + signal



import os
import time

ground_truth = "graph"
samples = "data"
test_folder = "test_pypi"

def test_manm_cs():
    os.system(
        "python3 -m pip install . &&" + # install dependencies from actual pip because they are not on testpypi
        f"mkdir -p {test_folder} && cd {test_folder} && rm -f {samples} {ground_truth} &&" +
        f"python3 -m manm_cs --num_nodes 10 --edge_density 0.5 --num_samples 10000 --discrete_node_ratio 0.5 --output_ground_truth_file \"{ground_truth}\" --output_samples_file \"{samples}\"")

    assert os.path.isfile(test_folder + "/" + ground_truth + ".gml")
    assert os.path.isfile(test_folder + "/" + samples + ".csv")



import numpy as np

from manm_cs.prob_distributions.continuous import ContinuousDistribution, GaussianDistribution


class BimodalDistribution(ContinuousDistribution):
    prob_dist1: GaussianDistribution
    prob_dist2: GaussianDistribution

    def __init__(self, prob_dist1: GaussianDistribution, prob_dist2: GaussianDistribution):
        super(BimodalDistribution, self).__init__()
        self.prob_dist1 = prob_dist1
        self.prob_dist2 = prob_dist2

    def sample(self, num_observations: int) -> np.array:
        num_obs1 = int(num_observations / 2)
        samples1 = self.prob_dist1.sample(num_observations=num_obs1)
        samples2 = self.prob_dist2.sample(num_observations=num_observations - num_obs1)

        samples = np.concatenate([samples1, samples2])
        np.random.shuffle(samples)

        return samples



import numpy as np

from manm_cs.prob_distributions.discrete.discrete_distribution import DiscreteDistribution


class UniformDiscreteDistribution(DiscreteDistribution):
    num_values: int

    def __init__(self, num_values: int):
        super(UniformDiscreteDistribution, self).__init__(num_values=num_values)

    def sample(self, num_observations: int) -> np.array:
        return np.random.randint(low=0, high=self.num_values - 1, size=num_observations)



from .binomial_distribution import BinomialDistribution
from .custom_discrete_distribution import CustomDiscreteDistribution
from .discrete_distribution import DiscreteDistribution
from .uniform_discrete_distribution import UniformDiscreteDistribution
from .random_discrete_distribution import RandomDiscreteDistribution



from manm_cs.noise import DiscreteNoise
from manm_cs.prob_distributions import CustomDiscreteDistribution


class DiscreteNoiseBuilder:
    signal_to_noise_ratio: float
    num_discrete_values: int

    def with_signal_to_noise_ratio(self, signal_to_noise_ratio: float) -> 'DiscreteNoiseBuilder':
        self.signal_to_noise_ratio = signal_to_noise_ratio
        return self

    def with_num_discrete_values(self, num_discrete_values: int) -> 'DiscreteNoiseBuilder':
        self.num_discrete_values = num_discrete_values
        return self

    def build(self) -> DiscreteNoise:
        signal = self.signal_to_noise_ratio
        num_noise_classes = self.num_discrete_values - 1
        avg_noise = (1 - signal) / num_noise_classes
        probs = [signal] + [avg_noise for _ in range(num_noise_classes)]

        return DiscreteNoise(prob_distribution=CustomDiscreteDistribution(probs=probs))



from typing import List

import numpy as np
from numpy.random import random_sample

from manm_cs.prob_distributions.discrete.discrete_distribution import DiscreteDistribution


class CustomDiscreteDistribution(DiscreteDistribution):
    probs: List[float]

    def __init__(self, probs: List[float]):
        if abs(sum(probs) - 1.0) > 0.001:
            raise ValueError(f'Expected the provided probabilities to approximately sum up to 1, '
                             f'but sum was {sum(probs)} '
                             f'[diff={abs(sum(probs) - 1.0)}, tolerance={0.001}]')
        self.probs = probs
        super(CustomDiscreteDistribution, self).__init__(num_values=len(self.probs))

    def sample(self, num_observations: int) -> np.array:
        bins = np.add.accumulate(self.probs)
        # example bins = [0.8,0.9,1.0]
        # index of interval random number is within 0 = [0,0.8] 1=[0.8,0.9] 2= [0.9,1]
        return np.digitize(random_sample(num_observations), bins)



import multiprocessing
from typing import List, Optional

import networkx as nx
import pandas as pd
import numpy as np
from pathos.multiprocessing import ProcessingPool

from manm_cs.variables.variable import Variable


class Graph:
    variables: List[Variable]

    def __init__(self, variables: List[Variable]):
        self.variables = variables

    def __get_top_sort_variables(self):
        # TODO: implement function
        return self.variables

    def sample(self, num_observations: int,
               num_processes: Optional[int] = None) -> List[pd.DataFrame]:
        """Return dataframe of size (num_observations, len(variables))

        """

        def fn(chunk_size: int):
            np.random.seed() # reseed thread to ensure independence

            df = pd.DataFrame()
            for variable in self.variables:
                df[variable.idx] = variable.sample(df=df, num_observations=chunk_size)
            return df

        if num_processes is None:
            max_num_processes = max(int(num_observations / 10000), 1)
            num_processes = min(multiprocessing.cpu_count(), max_num_processes)

        pool = ProcessingPool()
        chunk_sizes = [int(num_observations / num_processes) for _ in range(num_processes)]
        chunk_sizes[-1] = num_observations - sum(chunk_sizes[:-1])

        return pool.map(fn, chunk_sizes)

    def to_networkx_graph(self) -> nx.DiGraph:
        nx_graph = nx.DiGraph()
        for var in sorted(self.variables, key=lambda v: v.idx):
            nx_graph.add_node(var.idx)
            for parent in var.parents:
                nx_graph.add_edge(parent.idx, var.idx)
        return nx_graph



from manm_cs.noise import ContinuousNoise
from manm_cs.prob_distributions.continuous import GaussianDistribution


class GaussianNoiseBuilder:
    mu: float = .0
    sigma: float

    def with_sigma(self, sigma: float) -> 'GaussianNoiseBuilder':
        self.sigma = sigma
        return self

    def build(self) -> ContinuousNoise:
        prob_distribution = GaussianDistribution(mu=self.mu, sigma=self.sigma)
        return ContinuousNoise(prob_distribution=prob_distribution)



from abc import ABC
from typing import List, Callable, Optional

import pandas as pd

from manm_cs.noise.noise import Noise
from manm_cs.variables import VariableType


class Variable(ABC):
    idx: int
    type: VariableType
    parents: List['Variable']
    noise: Noise
    functions: List[Callable[..., float]]

    def __init__(self, idx: int, parents: List['Variable'], noise: Noise, functions: Optional[List[Callable[..., float]]] = None):
        self.idx = idx
        self.parents = parents
        self.noise = noise
        self.functions = functions

        # Input parameter validation
        if noise.get_type() != self.type:
            raise ValueError(f'Expected noise to be of type {self.type}, '
                             f'but was {noise.get_type()}')

    def _is_root(self) -> bool:
        return len(self.parents) == 0

    def _get_continuous_parents(self) -> List['Variable']:
        return list(filter(lambda p: p.type == VariableType.CONTINUOUS, self.parents))

    def _get_discrete_parents(self) -> List['Variable']:
        return list(filter(lambda p: p.type == VariableType.DISCRETE, self.parents))

    def sample(self, df: pd.DataFrame, num_observations: int) -> pd.Series:
        raise NotImplementedError()



import numpy as np

from manm_cs.prob_distributions.discrete.discrete_distribution import DiscreteDistribution


class BinomialDistribution(DiscreteDistribution):
    probability: float

    def __init__(self, probability: float):
        self.probability = probability
        super(BinomialDistribution, self).__init__(num_values=2)

    def sample(self, num_observations: int) -> np.array:
        return np.random.binomial(1, self.probability, num_observations)



from .continous_noise import ContinuousNoise
from .discrete_noise import DiscreteNoise
from .gaussian_noise_builder import GaussianNoiseBuilder
from .discrete_noise_builder import DiscreteNoiseBuilder



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from manm_cs.variables import VariableType

if TYPE_CHECKING:
    from manm_cs.prob_distributions.prob_distribution import ProbDistribution


class Noise:
    prob_distribution: 'ProbDistribution'

    def __init__(self, prob_distribution: 'ProbDistribution'):
        self.prob_distribution = prob_distribution

    def get_type(self) -> VariableType:
        return self.prob_distribution.type

    def __sample(self, num: int) -> np.array:
        return self.prob_distribution.sample(num_observations=num)

    def __add__(self, other) -> pd.Series:
        if not isinstance(other, pd.Series):
            raise ValueError(f'Cannot add other to Noise')

        return other.apply(lambda x: x + self.__sample(num=1)[0])



from typing import List, Dict, Tuple, Callable

import numpy as np
import pandas as pd

from manm_cs.noise.discrete_noise import DiscreteNoise
from manm_cs.prob_distributions import CustomDiscreteDistribution
from manm_cs.variables.variable import Variable, VariableType


class DiscreteVariable(Variable):
    type = VariableType.DISCRETE
    num_values: int
    mapping: Dict[Tuple[int, ...], int]

    def __init__(self, idx: int, num_values: int, noise: DiscreteNoise,
                 parents: List['DiscreteVariable'] = None):
        parents = [] if parents is None else parents
        super(DiscreteVariable, self).__init__(idx=idx, parents=parents, noise=noise)
        self.num_values = num_values
        self.logit_mapper_func = self.__create_logit_mapper_func()

        # Input parameter validation
        if self.noise.get_num_values() != self.num_values:
            raise ValueError(
                f'The noise term must define a probability distribution over all possible values. '
                f'Expected num_values equal to {self.num_values}, but received {self.noise.get_num_values()}')

    def mult_logit_function(self, value) -> pd.Series:
        categories =  list(range(0, self.num_values))

        e_sum = np.sum([np.exp(k*value) for  k in categories])
        softmax = [np.exp(k*value) / e_sum for  k in categories]
        if np.any(np.isnan(softmax)):
            ### check for overflow in np.exp, if large values are given tendency towards last category,
            ### hence we fix last category to 1
            softmax = [0 for k in categories]
            softmax[-1] = 1
        return softmax

    def __create_logit_mapper_func(self) -> Callable[[pd.Series], int]:
        """Creates the a multinomial logit function for each of the continuous parents

        """
        def func(parent_values: pd.Series):
            return np.sum([np.where(np.random.multinomial(1,self.mult_logit_function(value=value)) == 1)[0][0] for  value in parent_values])

        return func

    def get_non_root_signal(self, df: pd.DataFrame) -> pd.Series:
        # Compute signal for continuous parent variables
        continuous_parent_idxs = [p.idx for p in self._get_continuous_parents()]
        continuous_signal = df[continuous_parent_idxs].apply(self.logit_mapper_func, axis=1) \
            if len(continuous_parent_idxs) > 0 \
            else pd.Series(np.zeros(len(df)))

        # Compute signal for discrete parent variables
        discrete_parent_idxs = [p.idx for p in self._get_discrete_parents()]
        discrete_signal = df[discrete_parent_idxs].apply(sum, axis=1) \
            if len(discrete_parent_idxs) > 0 \
            else pd.Series(np.zeros(len(df)))

        # Aggregate continuous and discrete signal terms into overall signal term
        return (continuous_signal + discrete_signal).astype(int)

    def sample(self, df: pd.DataFrame, num_observations: int) -> pd.Series:
        if self._is_root():
            # If the variable is a root variable, the sampling is determined by the noise term only
            props = [1] + list(np.zeros(self.num_values - 1))
            signal = pd.Series(CustomDiscreteDistribution(props).sample(
                num_observations=num_observations)) # This will return a list of zeros
        else:
            # If the variable has one or more parent variables, the sampling is driven
            # by a combination of signal and noise term
            signal = self.get_non_root_signal(df=df)

        # Combine noise and signal terms and apply ring transformation
        return (self.noise + signal) % self.num_values



from manm_cs.graph.graph import Graph
from manm_cs.noise import DiscreteNoise
from manm_cs.prob_distributions import BinomialDistribution, CustomDiscreteDistribution, UniformDiscreteDistribution
from manm_cs.variables.discrete_variable import DiscreteVariable
import networkx as nx


def create_simple_discrete_model2():
    A = DiscreteVariable(idx=0, num_values=2, noise=DiscreteNoise(prob_distribution=BinomialDistribution(probability=.25)))
    B = DiscreteVariable(idx=1, num_values=3, noise=DiscreteNoise(UniformDiscreteDistribution(num_values=3)))
    C = DiscreteVariable(idx=2, num_values=3, parents=[A, B],
                         noise=DiscreteNoise(
                             prob_distribution=CustomDiscreteDistribution(probs=[.5, .2, .3])))
    variables = [A, B, C]
    return Graph(variables=variables)


def test_to_networkx_graph():
    expected = nx.DiGraph()
    expected.add_node(0)
    expected.add_node(1)
    expected.add_node(2)
    expected.add_edge(0, 2)
    expected.add_edge(1, 2)

    graph = create_simple_discrete_model2()
    actual = graph.to_networkx_graph()

    assert nx.is_isomorphic(expected, actual)



import numpy as np

from manm_cs.prob_distributions.continuous import ContinuousDistribution


class GaussianDistribution(ContinuousDistribution):
    mu: float
    sigma: float

    def __init__(self, mu: float, sigma: float):
        super(GaussianDistribution, self).__init__()
        self.mu = mu
        self.sigma = sigma

    def sample(self, num_observations: int) -> np.array:
        return np.random.normal(loc=self.mu, scale=self.sigma, size=num_observations)



from typing import List

import pandas as pd


def write_single_csv(dataframes: List[pd.DataFrame], target_path: str):
    dataframes[0].to_csv(target_path, index=False)
    for df in dataframes[1:]:
        df.to_csv(target_path, mode='a', header=False, index=False)


def write_single_hdf(dataframes: List[pd.DataFrame], target_path: str):
    dataframes[0].to_hdf(target_path, key='generated_dataset', format='table')
    for df in dataframes[1:]:
        df.to_hdf(target_path, "generated_dataset", format='table', append=True)



from typing import Dict, Optional, Callable, List, Tuple

import networkx as nx
import numpy as np
import random
from validation import validate_int, validate_float, validate_bool, validate_text

from manm_cs.graph import Graph
from manm_cs.noise import GaussianNoiseBuilder, DiscreteNoiseBuilder
from manm_cs.prob_distributions.continuous.bimodal_distribution import BimodalDistribution
from manm_cs.prob_distributions.continuous.gaussian_distribution import GaussianDistribution
from manm_cs.variables.continuous_variable import ContinuousVariable
from manm_cs.variables.discrete_variable import DiscreteVariable
from manm_cs.variables.variable import Variable
from manm_cs.variables.variable_type import VariableType


class GraphBuilder:
    num_nodes: int
    edge_density: float
    discrete_node_ratio: float
    conditional_gaussian: bool
    num_samples: int

    discrete_signal_to_noise_ratio: float
    min_discrete_value_classes: Optional[int] = None
    max_discrete_value_classes: Optional[int] = None
    continuous_noise_std: float

    beta_lower_limit: float
    beta_upper_limit: float

    functions: List[Tuple[float, Callable[...,float]]]

    dag: Optional[nx.DiGraph] = None

    def with_graph_structure_file(self, file_name: str) -> 'GraphBuilder':
        validate_text(file_name, pattern='.*\.gml')
        self.graph_structure_file_name = file_name
        return self

    def with_networkx_DiGraph(self, graph: 'DiGraph') -> 'GraphBuilder':
        assert nx.is_directed_acyclic_graph(graph)
        self.with_num_nodes(nx.number_of_nodes(graph))
        self.with_edge_density(nx.density(graph))
        self.dag = graph
        return self

    def with_num_nodes(self, num_nodes: int) -> 'GraphBuilder':
        validate_int(num_nodes, min_value=1)
        self.num_nodes = num_nodes
        return self

    def with_edge_density(self, edge_density: float) -> 'GraphBuilder':
        validate_float(edge_density, min_value=0.0, max_value=1.0)
        self.edge_density = edge_density
        return self

    def with_discrete_node_ratio(self, discrete_node_ratio: float) -> 'GraphBuilder':
        validate_float(discrete_node_ratio, min_value=0.0, max_value=1.0)
        self.discrete_node_ratio = discrete_node_ratio
        return self

    def with_conditional_gaussian(self, conditional_gaussian: bool) -> 'GraphBuilder':
        validate_bool(conditional_gaussian)
        self.conditional_gaussian = conditional_gaussian
        return self

    def with_num_samples(self, num_samples: int) -> 'GraphBuilder':
        validate_int(num_samples, min_value=1)
        self.num_samples = num_samples
        return self

    def with_discrete_signal_to_noise_ratio(self,
                                            discrete_signal_to_noise_ratio: float) -> 'GraphBuilder':
        validate_float(discrete_signal_to_noise_ratio, min_value=0.0, max_value=1.0)
        self.discrete_signal_to_noise_ratio = discrete_signal_to_noise_ratio
        return self

    def with_min_discrete_value_classes(self, min_discrete_value_classes: int) -> 'GraphBuilder':
        validate_int(min_discrete_value_classes, min_value=1)
        if self.max_discrete_value_classes is not None \
                and min_discrete_value_classes > self.max_discrete_value_classes:
            raise ValueError(f'Expected min_discrete_value_classes to be smaller ' +
                             'or equal to max_discrete_value_classes, but are '
                             f'{min_discrete_value_classes} and {self.max_discrete_value_classes}')
        self.min_discrete_value_classes = min_discrete_value_classes
        return self

    def with_max_discrete_value_classes(self, max_discrete_value_classes: int) -> 'GraphBuilder':
        validate_int(max_discrete_value_classes, min_value=1)
        if self.min_discrete_value_classes is not None \
                and self.min_discrete_value_classes > max_discrete_value_classes:
            raise ValueError(f'Expected max_discrete_value_classes to be greater ' +
                             'or equal to min_discrete_value_classes, but are ' +
                             f'{max_discrete_value_classes} and {self.min_discrete_value_classes}')
        self.max_discrete_value_classes = max_discrete_value_classes
        return self

    def with_continuous_noise_std(self, continuous_noise_std: float) -> 'GraphBuilder':
        validate_float(continuous_noise_std, min_value=0.0)
        self.continuous_noise_std = continuous_noise_std
        return self

    def with_functions(self, function_tuples: List[Tuple[float, Callable[...,float]]]) -> 'GraphBuilder':
        ### Transform input in form of
        # [(0.5 F1), (0.3 F2), (0.2 F3)]
        # into, to allow drawing later on using random number between 0 and 1
        # [(0.5 F1), (0.8 F2), (1.0 F3)]
        #TODO validation
        self.functions = []
        cur_range = 0.0
        for function_tuple in function_tuples:
            cur_range += function_tuple[0]
            if cur_range > 1.0:
                raise ValueError(f'Ranges of functions should not exceeed 1.0')
            self.functions.append((cur_range,function_tuple[1]))
        if cur_range < 1.0:
            raise ValueError(f'Ranges of functions should sum up to 1.0 but are ' +
                             f'{cur_range}')
        return self

    def with_betas(self, lower_limit: float, upper_limit: float) -> 'GraphBuilder':
        # Switch limits if lower is larger than upper
        if lower_limit > upper_limit:
            tmp = upper_limit
            upper_limit = lower_limit
            lower_limit = tmp
        validate_float(lower_limit, min_value=0.0)
        if lower_limit <= 0.0:
            raise ValueError(f'Range of beta has to be within (0, Inf), ' +
                             f'but is {lower_limit} and {upper_limit}')
        validate_float(upper_limit, min_value=lower_limit)
        self.beta_lower_limit = lower_limit
        self.beta_upper_limit = upper_limit
        return self

    def chose_function(self):
        rand_val = random.random()
        for function_tuple in sorted(self.functions):
            if rand_val <= function_tuple[0]:
                return function_tuple[1]
        ### use last entry as default
        def identical(value):
            return value
        return identical

    def sample_beta(self) -> List[float]:
        if np.random.randint(0,2) == 0:
            # we sample beta from [-upper, -lower]
            return random.uniform(-self.beta_upper_limit, -self.beta_lower_limit)
        else:
            # we sample beta from [lower, upper]
            return random.uniform(self.beta_lower_limit, self.beta_upper_limit)

    def generate_discrete_variable(self, parents, node_idx) -> 'DiscreteVariable':
        num_values = np.random.randint(
            low=self.min_discrete_value_classes,
            high=self.max_discrete_value_classes + 1,
            size=1)[0]
        noise = DiscreteNoiseBuilder() \
            .with_signal_to_noise_ratio(
            signal_to_noise_ratio=self.discrete_signal_to_noise_ratio) \
            .with_num_discrete_values(num_discrete_values=num_values) \
            .build()
        return DiscreteVariable(idx=node_idx, num_values=num_values,
                                parents=parents, noise=noise)

    def generate_continuous_variable(self, parents, node_idx) -> 'ContinuousVariable':
        noise = GaussianNoiseBuilder() \
            .with_sigma(sigma=self.continuous_noise_std) \
            .build()
        num_continuous_parents = sum(
            [1 for p in parents if p.type == VariableType.CONTINUOUS])

        functions = [self.chose_function() for p in range(num_continuous_parents)]
        betas = [self.sample_beta() for p in range(num_continuous_parents)]
        return ContinuousVariable(idx=node_idx, parents=parents, functions=functions,
                                  noise=noise, betas=betas)

    def generate_dag(self, seed: int) -> 'DiGraph':
        # Generate graph using networkx package
        G = nx.gnp_random_graph(n=self.num_nodes, p=self.edge_density, seed=seed, directed=True)
        # Convert generated graph to DAG
        dag = nx.DiGraph()
        dag.add_nodes_from(G)
        dag.add_edges_from([(u, v, {}) for (u, v) in G.edges() if u < v])
        assert nx.is_directed_acyclic_graph(dag)
        return dag

    def build(self, seed: int = 0) -> Graph:
        self.dag = self.generate_dag(seed) if self.dag is None else self.dag

        # Create list of topologically sorted nodes
        # Note, nodes are ordered already, sorting step may become relevant, if graph generation above is changed
        top_sort_idx = list(nx.topological_sort(self.dag))
        num_discrete_nodes = int(self.discrete_node_ratio * self.num_nodes)

        variables_by_idx: Dict[int, Variable] = {}
        for i, node_idx in enumerate(top_sort_idx):
            parents = [variables_by_idx[idx] for idx in sorted(list(self.dag.predecessors(node_idx)))]

            # Conditional Gaussian:
            if self.conditional_gaussian == True:
                if i < num_discrete_nodes:
                    # Consider the first num_discrete_nodes nodes to be of discrete type
                    variable = self.generate_discrete_variable(parents, node_idx)
                else:
                    variable = self.generate_continuous_variable(parents, node_idx)
            # Mixed:
            else:
                # For each node: decide for discrete or conintuous given probability of self.discrete_node_ratio
                # Discrete:
                if random.random() < self.discrete_node_ratio:
                    variable = self.generate_discrete_variable(parents, node_idx)
                # Continuous:
                else:
                    variable = self.generate_continuous_variable(parents, node_idx)

            variables_by_idx[node_idx] = variable

        variables = [variables_by_idx[idx] for idx in top_sort_idx]
        return Graph(variables=variables)



import numpy as np

from manm_cs.prob_distributions.discrete import CustomDiscreteDistribution
from manm_cs.prob_distributions.discrete import DiscreteDistribution


class RandomDiscreteDistribution(DiscreteDistribution):
    distribution: CustomDiscreteDistribution

    def __init__(self, num_values: int):
        super(RandomDiscreteDistribution, self).__init__(num_values=num_values)
        self.distribution = self.__generate_prob_distribution()

    def __generate_prob_distribution(self):
        probs = []
        for i in range(self.num_values - 1):
            next_prob = np.random.uniform(0, 1 - sum(probs))
            probs.append(next_prob)
        probs.append(1 - sum(probs))

        return CustomDiscreteDistribution(probs=probs)

    def sample(self, num_observations: int) -> np.array:
        return self.distribution.sample(num_observations=num_observations)



from abc import ABC

from manm_cs.prob_distributions.prob_distribution import ProbDistribution
from manm_cs.variables import VariableType


class DiscreteDistribution(ProbDistribution, ABC):
    type = VariableType.DISCRETE
    num_values: int

    def __init__(self, num_values: int):
        self.num_values = num_values

    def get_num_values(self) -> int:
        return self.num_values



from abc import ABC

import numpy as np

from manm_cs.variables import VariableType


class ProbDistribution(ABC):
    type: VariableType

    def sample(self, num_observations: int) -> np.array:
        raise NotImplementedError()



from manm_cs.noise.noise import Noise
from manm_cs.prob_distributions.continuous import ContinuousDistribution


class ContinuousNoise(Noise):

    def __init__(self, prob_distribution: ContinuousDistribution):
        super(ContinuousNoise, self).__init__(prob_distribution=prob_distribution)



from manm_cs.prob_distributions.continuous import GaussianDistribution
from manm_cs.prob_distributions.discrete import BinomialDistribution
from manm_cs.prob_distributions.discrete import CustomDiscreteDistribution
from manm_cs.prob_distributions.discrete import UniformDiscreteDistribution
